//==================================================================
//
//	File name	:	frame.js
//
//	Summary		:	Window frame class
//
//	Ceated		:	2018/11/20
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

(function (frame) {
	//==========================================================
	//	Global
	//==========================================================
	// Const
	frame.STYLE_TITLEBAR = 1;
	frame.STYLE_CLOSEBUTTON = 2;
	frame.STYLE_DESTROY_ON_CLOSE = 4;
	frame.STYLE_CLIENTEDGE = 8;
	frame.STYLE_RESIZABLE = 16;
	frame.STYLE_MOBABLE = 32;
	frame.STYLE_SCROLLABLE = 64;
	frame.STYLE_IMAGE_ICON = 128;
	frame.STYLE_CSS_ICON = 256;
	frame.RESIZE_NONE = 0;
	frame.RESIZE_LEFT = 1;
	frame.RESIZE_RIGHT = 2;
	frame.RESIZE_TOP = 3;
	frame.RESIZE_BOTTOM = 4;
	frame.RESIZE_BOTH = 5;
	frame.CENTER_POSITION = -999999;
	
	// Grobal variables
	frame.windows = {};
	frame.windowCount = 0;
	
	// Private variables
	let iResizeMode = frame.RESIZE_NONE;
	let iMovingX = 0;
	let iMovingY = 0;
	let currentModalLevel = 0;
	
	
	//==========================================================
	//
	//	Class		:	Window
	//
	//	Summary		:	Class of window
	//
	//==========================================================
	frame.Window = class extends jaswin.BaseWindow
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[String]
					oRect,			// * Size and position[ExpandableRect]
					iMinWidth,		// * Minimum width[int]
					iMinHeight,		// * Minimum height[int]
					iStyle,			// * Window style[int:bit flag]
					sTitle,			// Title[String]
					sIcon,			// Icon URL or icon class[String]
					bHideOnCreate)	// Hide on create or not[boolean]
		{
			super(sID, oRect);
			
			if((document.getElementById(sID)) || (frame.windows[sID]))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_DUPLICATION_ID, sID),
											"frame.js", "jaswin.frame.Window.constructor");
			
			if(typeof iMinWidth != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_MIN_WIDTH, sID),
											"frame.js", "jaswin.frame.Window.constructor");
			
			if(iMinWidth < 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_MIN_WIDTH, sID),
											"frame.js", "jaswin.frame.Window.constructor");
			
			if(typeof iMinHeight != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_MIN_HEIGHT, sID),
											"frame.js", "jaswin.frame.Window.constructor");
			
			if(iMinHeight < 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_MIN_HEIGHT, sID),
											"frame.js", "jaswin.frame.Window.constructor");
			
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"frame.js", "jaswin.frame.Window.constructor");
			
			if(sTitle)
			{
				if(typeof sTitle != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TITLE, sID),
												"frame.js", "jaswin.frame.Window.constructor");
			}
			
			if(sIcon)
			{
				if(typeof sIcon != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ICON_URL, sID),
												"frame.js", "jaswin.frame.Window.constructor");
			}
			
			if(typeof bHideOnCreate != "undefined")
			{
				if(typeof bHideOnCreate != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_HIDE_ON_CREATE, sID),
												"frame.js", "jaswin.frame.Window.constructor");
			}
			
			// Set properties
			this.minWidth = iMinWidth;			// Minimum width[int]
			this.minHeight = iMinHeight;		// Minimum height[int]
			this.style = iStyle;				// Window style[int:bit flag]
			this.center = null;					// Center element[DOM Element]
			this.titleBar = null;				// Title bar element[DOM Element]
			this.icon = null;					// Icon image element[DOM Element]
			this.titleSpan = null;				// Title span element[DOM Element]
			this.closeButton = null;			// Close button element[DOM Element]
			this.zIndex = 0;					// Z-index[int]
			this.modalLevel = currentModalLevel	// Modal level[int]
			this.modalPanel = null;				// Modal base panel element[DOM Element]
			this.returnValue = null;			// Return value[any]
			this.returnWindow = null;			// Return window[Window]
			this.callBack = null;				// Call back function[function]
			
			// Create element
			this.element = document.createElement("div");
			this.element.id = this.id;
			this.element.className = "window";
			
			// Horizontal rect
			if(this.rect.anchorX == jaswin.ANCHOR_BOTH)
			{
				this.element.style.left = this.rect.left + "px";
				this.element.style.right = this.rect.right + "px";
			}
			else
			{
				if(this.rect.width < this.minWidth)
					this.rect.width = this.minWidth;
				
				if(this.rect.anchorX == jaswin.ANCHOR_LEFT)
					this.element.style.left = this.rect.left + "px";
				else if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
					this.element.style.right = this.rect.right + "px";
				this.element.style.width = this.rect.width + "px";
			}
			
			// Vertical rect
			if(this.rect.anchorY == jaswin.ANCHOR_BOTH)
			{
				this.element.style.top = this.rect.top + "px";
				this.element.style.bottom = this.rect.bottom + "px";
			}
			else
			{
				if(this.rect.height < this.iMinHeight)
					this.rect.height = this.iMinHeight;
				
				if(this.rect.anchorY == jaswin.ANCHOR_TOP)
					this.element.style.top = this.rect.top + "px";
				else if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
					this.element.style.bottom = this.rect.bottom + "px";
				this.element.style.height = this.rect.height + "px";
			}
			
			// Top Edge
			let oTopEdge = document.createElement("div");
			oTopEdge.className = "edge";
			oTopEdge.style.height = "2px";
			oTopEdge.style.top = "0px";
			oTopEdge.style.left = "0px";
			oTopEdge.style.right = "0px";
			this.element.appendChild(oTopEdge);
			
			// Left Edge
			let oLeftEdge = document.createElement("div");
			oLeftEdge.className = "edge";
			oLeftEdge.style.left = "0px";
			oLeftEdge.style.top = "2px";
			oLeftEdge.style.bottom = "2px";
			oLeftEdge.style.width = "2px";
			this.element.appendChild(oLeftEdge);
			
			// Center(For client area)
			this.center = document.createElement("div");
			this.center.style.position = "absolute";
			this.center.style.left = "2px";
			this.center.style.right = "2px";
			this.center.style.top = "2px";
			this.center.style.bottom = "2px";
			this.element.appendChild(this.center);
			
			// Right edge
			let oRightEdge = document.createElement("div");
			oRightEdge.className = "edge";
			oRightEdge.style.right = "0px";
			oRightEdge.style.top = "2px";
			oRightEdge.style.bottom = "2px";
			oRightEdge.style.width = "2px";
			this.element.appendChild(oRightEdge);
			
			// Bottom edge
			let oBottomEdge = document.createElement("div");
			oBottomEdge.className = "edge";
			oBottomEdge.style.left = "0px";
			oBottomEdge.style.right = "2px";
			oBottomEdge.style.bottom = "0px";
			oBottomEdge.style.height = "2px";
			this.element.appendChild(oBottomEdge);
			
			// Right bottom corner edge
			let oCornerEdge = document.createElement("div");
			oCornerEdge.className = "edge";
			oCornerEdge.style.right = "0px";
			oCornerEdge.style.bottom = "0px";
			oCornerEdge.style.width = "2px";
			oCornerEdge.style.height = "2px";
			this.element.appendChild(oCornerEdge);
			
			// Title bar
			if(this.style & frame.STYLE_TITLEBAR)
			{
				this.titleBar = document.createElement("div");
				this.titleBar.className = "title";
				
				if(sIcon)
				{
					if(this.style & frame.STYLE_IMAGE_ICON)
					{
						// Image icon
						this.icon = document.createElement("img");
						this.icon.src = sIcon;
						this.icon.width = 16;
						this.icon.height = 16;
						this.icon.style.position = "absolute";
						this.icon.style.left = "1px";
						this.icon.style.top = "1px";
						this.titleBar.appendChild(this.icon);
					}
					else if(this.style & frame.STYLE_CSS_ICON)
					{
						// Span icon
						this.icon = document.createElement("span");
						this.icon.style.position = "absolute";
						this.icon.style.width = "16px";
						this.icon.style.height = "16px";
						this.icon.style.left = "1px";
						this.icon.style.top = "1px";
						
						let oSpan = document.createElement("span");
						oSpan.className = sIcon;
						this.icon.appendChild(oSpan);
						this.titleBar.appendChild(this.icon);
					}
				}
				
				this.titleSpan = document.createElement("span");
				this.titleSpan.className = "title-text";
				if(sTitle)
					this.titleSpan.innerHTML = sTitle;
				
				if(sIcon)
					this.titleSpan.style.left = "22px";
				else
					this.titleSpan.style.left = "4px";
				
				if(this.style & frame.STYLE_CLOSEBUTTON)
					this.titleSpan.style.right = "36px";
				else
					this.titleSpan.style.right = "4px";
				this.titleBar.appendChild(this.titleSpan);
				
				if(this.style & frame.STYLE_CLOSEBUTTON)
				{
					// Close button
					this.closeButton = document.createElement("span");
					this.closeButton.className = "close";
					this.closeButton.innerHTML = "x";
					this.closeButton.addEventListener("click", this.closeEvent = (e) => {
						this.close();
					}, false);
					
					this.titleBar.appendChild(this.closeButton);
				}
				this.center.appendChild(this.titleBar);
			}
			
			// Client area
			this.container = document.createElement("div");
			
			if(this.style & frame.STYLE_CLIENTEDGE)
				this.container.className = "clientedge";
			else
				this.container.className = "clieantflat";
			
			if(this.style & frame.STYLE_SCROLLABLE)
				this.container.style.overflow = "auto";
			
			if(this.style & frame.STYLE_TITLEBAR)
				this.container.style.top = "18px";
			else
				this.container.style.top = "0px";
			this.center.appendChild(this.container);
			
			// Window forcus event
			this.element.addEventListener("mousedown", this.selectWindowEvent = (e) => {
				this.setActive();
				e.stopPropagation();
			}, false);
			
			if(this.style & frame.STYLE_RESIZABLE)
			{
				// Resize event
				if(this.rect.anchorX != jaswin.ANCHOR_BOTH)
				{
					oLeftEdge.addEventListener("mousedown", startResizeWindow, false);
					oLeftEdge.style.cursor = "w-resize";
					oRightEdge.addEventListener("mousedown", startResizeWindow, false);
					oRightEdge.style.cursor = "e-resize";
				}
				if(this.rect.anchorY != jaswin.ANCHOR_BOTH)
				{
					oTopEdge.addEventListener("mousedown", startResizeWindow, false);
					oTopEdge.style.cursor = "n-resize";
					oBottomEdge.addEventListener("mousedown", startResizeWindow, false);
					oBottomEdge.style.cursor = "s-resize";
				}
				
				if((this.rect.anchorX != jaswin.ANCHOR_BOTH)
				&& (this.rect.anchorY != jaswin.ANCHOR_BOTH))
				{
					oCornerEdge.addEventListener("mousedown", startResizeWindow, false);
					oCornerEdge.style.cursor = "se-resize";
				}
			}
			
			if(this.style & frame.STYLE_MOBABLE)
			{
				// Move event
				if((this.rect.anchorX == jaswin.ANCHOR_LEFT)
				&& (this.rect.anchorY == jaswin.ANCHOR_TOP))
					this.titleBar.addEventListener("mousedown", startDragWindow, false);
			}
			
			frame.windows[this.id] = this;
			frame.windowCount++;
			
			if((typeof bHideOnCreate != "undefined") && bHideOnCreate)
			{
				for(let sKey in frame.windows)
				{
					if(frame.windows[sKey] == this)
					{
						this.zIndex = 1;
						this.element.style.zIndex = 1 + this.modalLevel;
						this.element.style.display = "none";
					}
					else
					{
						frame.windows[sKey].zIndex++;
						frame.windows[sKey].element.style.zIndex = parseInt(frame.windows[sKey].element.style.zIndex) + 1;
					}
				}
			}
			else
			{
				this.zIndex = frame.windowCount;
				this.element.style.zIndex = frame.windowCount + this.modalLevel;
			}
		}
		
		//==================================================
		//	Delete DOM element
		//==================================================
		destroy()
		{
			let iCurIndex = 0;
			let iTmpIndex = 0;
			
			iCurIndex = this.zIndex;
			for(let sKey in frame.windows)
			{
				if(frame.windows[sKey].modalLevel == this.modalLevel)
				{
					iTmpIndex = frame.windows[sKey].zIndex;
					if((frame.windows[sKey] != this) && (frame.windows[sKey].zIndex > iCurIndex))
					{
						iTmpIndex--;
						frame.windows[sKey].zIndex = iTmpIndex;
						frame.windows[sKey].element.style.zIndex = iTmpIndex + frame.windows[sKey].modalLevel;
					}
				}
			}
			super.destroy();
			delete frame.windows[this.id];
			frame.windowCount--;
			
			this.parent.notifyDestroyWindow(this);
		}
		
		//==================================================
		//	Set active
		//==================================================
		setActive()
		{
			if(this.modalLevel != currentModalLevel)
				return;
			
			super.setActive();
			
			if(this.style & frame.STYLE_TITLEBAR)
			{
				this.titleBar.className = "title-active";
				this.titleSpan.className = "title-text-active";
			}
			
			if(!this.windowKeyEvent)
			{
				this.element.addEventListener("keydown", this.windowKeyEvent = (e) => {
					if(jaswin.currentWindow != this)
						return true;
					
					if(e.keyCode == 9)
					{
						// Tab
						if(e.shiftKey)
							return this.prevTabstop(e);
						else
							return this.nextTabstop(e);
					}
					return true;
				}, false);
			}
		}
		
		//==================================================
		//	Set inctive
		//==================================================
		setInactive()
		{
			super.setInactive();
			
			if(this.style & frame.STYLE_TITLEBAR)
			{
				this.titleBar.className = "title";
				this.titleSpan.className = "title-text";
			}
			
			if(this.windowKeyEvent)
			{
				this.element.removeEventListener("keydown", this.windowKeyEvent, false);
				this.windowKeyEvent = null;
			}
		}
		
		//==================================================
		//	Show this window
		//==================================================
		show()
		{
			super.show();
			this.setActive();
		}
		
		//==================================================
		//	Hide this control
		//==================================================
		hide()
		{
			let iCurIndex = 0;
			let iMaxIndex = 0;
			let iTmpIndex = 0;
			let oWnd = null;
			
			iCurIndex = this.zIndex;
			for(let sKey in frame.windows)
			{
				if(frame.windows[sKey].modalLevel == this.modalLevel)
				{
					if(frame.windows[sKey] == this)
					{
						this.element.style.zIndex = 1;
						this.setInactive();
					}
					else
					{
						iTmpIndex = frame.windows[sKey].zIndex;
						if(frame.windows[sKey].zIndex < iCurIndex)
						{
							iTmpIndex++;
							frame.windows[sKey].zIndex = iTmpIndex;
							frame.windows[sKey].element.style.zIndex = iTmpIndex + frame.windows[sKey].modalLevel;
						}
						
						if(iTmpIndex > iMaxIndex)
						{
							oWnd = frame.windows[sKey];
							iMaxIndex = iTmpIndex;
						}
					}
				}
			}
			super.hide();
			
			if(oWnd)
				oWnd.setActive();
		}
		
		//==================================================
		//	Close this window
		//==================================================
		close()
		{
			if(this.modalPanel)
			{
				global.element.removeChild(this.modalPanel);
				currentModalLevel -= 1000;
			}
			this.hide();
			
			if(this.returnWindow)
				this.returnWindow.setActive();
			
			if(this.callBack)
				this.callBack(this);
			
			if(this.style & frame.STYLE_DESTROY_ON_CLOSE)
				this.destroy();
		}
		
		//==================================================
		//	Show this window as dialog box
		//==================================================
		showDialog(oWindow,			// Return window on close[Window class]
					fpCallBack)		// Call back function[function]
		{
			if(oWindow)
			{
				if((typeof oWindow != "object")
				|| (!oWindow instanceof frame.Window))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_WINDOW, this.id),
												"frame.js", "jaswin.frame.Window.showDialog");
				
				this.returnWindow = oWindow;
			}
			
			if(fpCallBack)
			{
				if(typeof fpCallBack != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_CALLBACK, this.id),
												"frame.js", "jaswin.frame.Window.showDialog");
				this.callBack = fpCallBack;
			}
			
			currentModalLevel += 1000;
			this.modalPanel = document.createElement("div");
			this.modalPanel.className = "modal-panel";
			this.modalPanel.style.zIndex = currentModalLevel;
			global.element.appendChild(this.modalPanel);
			this.modalLevel = currentModalLevel;
			this.show();
		}
		
		//==================================================
		//	Change size of this window
		//==================================================
		setSize(iWidth,		// Width to set
				iHeight)	// Height to set
		{
			if(typeof iWidth != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WIDTH, this.id),
											"frame.js", "jaswin.frame.Window.setSize");
			
			if(typeof iHeight != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_HEIGHT, this.id),
											"frame.js", "jaswin.frame.Window.setSize");
			
			if(this.rect.anchorX != jaswin.ANCHOR_BOTH)
				this.rect.width = iWidth;
			if(this.rect.anchorY != jaswin.ANCHOR_BOTH)
				this.rect.height = iHeight;
			
			this.element.style.width = iWidth + "px";
			this.element.style.height = iHeight + "px";
			
			if((this.toolbar) && (this.toolbar.isVisible()))
			{
				let iTop = 0;
				let iOffset = 0;
				
				if(this.style & frame.STYLE_TITLEBAR)
					iTop += 18;
				if(this.menu)
					iTop += 22;
				
				if(this.toolbar.controls.length > 0)
					iOffset = this.toolbar.controls[this.toolbar.controls.length - 1].element.offsetTop;
				
				this.toolbar.element.style.height = 20 + iOffset + "px";
				iTop += 22 + iOffset;
				
				this.container.style.top = iTop + "px";
			}
			
			if(this.onResize)
				this.onResize();
		}
		
		//==================================================
		//	Move this window
		//==================================================
		setPosition(iX,		// Horizontal position
					iY)		// Vertical position
		{
			if(typeof iX != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_X, this.id),
											"frame.js", "jaswin.frame.Window.setPosition");
			
			if(typeof iY != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_Y, this.id),
											"frame.js", "jaswin.frame.Window.setPosition");
			
			this.rect.left = iX;
			this.rect.top = iY;
			
			this.element.style.left = iX + "px";
			this.element.style.top = iY + "px";
		}
		
		//==================================================
		//	Set this control to parent class
		//==================================================
		setParent(oParent)
		{
			if((typeof oParent != "object")
			|| (!oParent instanceof jaswin.Container)
			|| (oParent instanceof jaswin.frame.Window))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_PARENT, this.id),
											"frame.js", "jaswin.frame.Window.setParent");
			
			this.parent = oParent;
			
			if((this.rect.anchorX == jaswin.ANCHOR_LEFT)
			&& (this.rect.anchorY == jaswin.ANCHOR_TOP))
			{
				if(this.rect.left == frame.CENTER_POSITION)
				{
					let iParentWidth = this.parent.container.clientWidth;
					
					if(this.rect.width >= iParentWidth)
						this.rect.left = 0;
					else
						this.rect.left = Math.floor((iParentWidth - this.rect.width) / 2);
					
					this.element.style.left = this.rect.left + "px";
				}
				if(this.rect.top == frame.CENTER_POSITION)
				{
					let iParentHeight = this.parent.container.clientHeight;
					
					if(this.rect.height >= iParentHeight)
						this.rect.top = 0;
					else
						this.rect.top = Math.floor((iParentHeight - this.rect.height) / 2);
					
					this.element.style.top = this.rect.top + "px";
				}
			}
			
			global[this.id] = this;
		}
		
		//==================================================
		//	Set main menu
		//==================================================
		setMenu(oMenuBar)
		{
			super.setMenu(oMenuBar);
			
			this.center.appendChild(this.menu.element);
			this.menu.setParent(this);
			
			let iTop = 22;
			
			if(this.style & frame.STYLE_TITLEBAR)
			{
				iTop += 18;
				this.menu.element.style.top = "18px";
			}
			
			if(this.toolbar)
				iTop += parseInt(this.toolbar.style.height) + 2;
			
			this.container.style.top = iTop + "px";
			
			this.element.addEventListener("menu", (e) => {
				let iKeyCode = 0;
				let sChar = "";
				
				if(!this.menu) return;
				
				iKeyCode = e.detail;
				sChar = String.fromCharCode(iKeyCode);
				this.menu.shortCut(sChar);
			}, false);
		}
		
		//==================================================
		//	Set toolbar
		//==================================================
		setToolBar(oToolBar)
		{
			super.setToolBar(oToolBar);
			
			let iTop = 0;
			let iOffset = 0;
			
			if(this.style & frame.STYLE_TITLEBAR)
				iTop += 18;
			if(this.menu)
				iTop += 22;
			
			this.toolbar.element.style.top = iTop + "px";
			this.center.appendChild(this.toolbar.element);
			this.toolbar.setParent(this);
			
			if(this.toolbar.controls.length > 0)
				iOffset = this.toolbar.controls[this.toolbar.controls.length - 1].element.offsetTop;
			
			this.toolbar.element.style.height = 20 + iOffset + "px";
			iTop += 22 + iOffset;
			
			this.container.style.top = iTop + "px";
		}
		
		//==================================================
		//	Show tool bar
		//==================================================
		showToolBar()
		{
			if(!this.toolbar)
				return;
			this.toolbar.show();
			
			let iTop = 0;
			let iOffset = 0;
			
			if(this.style & frame.STYLE_TITLEBAR)
				iTop += 18;
			if(this.menu)
				iTop += 22;
			
			if(this.toolbar.controls.length > 0)
				iOffset = this.toolbar.controls[this.toolbar.controls.length - 1].element.offsetTop;
			
			this.toolbar.element.style.height = 20 + iOffset + "px";
			iTop += 22 + iOffset;
			
			this.container.style.top = iTop + "px";
		}
		
		//==================================================
		//	Hide tool bar
		//==================================================
		hideToolBar()
		{
			if(!this.toolbar)
				return;
			this.toolbar.hide();
			
			let iTop = 0;
			
			if(this.style & frame.STYLE_TITLEBAR)
				iTop += 18;
			if(this.menu)
				iTop += 22;
			
			this.container.style.top = iTop + "px";
		}
		
		//==================================================
		//	Set context menu to this window
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			super.setContextMenu(oMenu, fpContext);
			
			if(!this.windowContextMenu)
			{
				this.container.addEventListener("contextmenu", this.windowContextMenu = (e) => {
					this.showContextMenu(e);
					return false;
				}, false);
			}
		}
		
		//==================================================
		//	Remove context menu from this window
		//==================================================
		removeContextMenu()
		{
			super.removeContextMenu();
			if(this.windowContextMenu)
				this.container.removeEventListener("contextmenu", this.windowContextMenu, false);
			this.windowContextMenu = null;
		}
		
		//==================================================
		//	Set title
		//==================================================
		setTitle(sTitle)
		{
			if(this.style & frame.STYLE_TITLEBAR)
			{
				if(typeof sTitle != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TITLE, this.id),
												"frame.js", "jaswin.frame.Window.setTitle");
				this.titleSpan.innerHTML = sTitle;
			}
		}
		
		//==================================================
		//	Set call back function on close
		//==================================================
		setCallback(fpCallBack)
		{
			if(typeof fpCallBack != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CALLBACK, this.id),
											"frame.js", "jaswin.frame.Window.setCallback");
			this.callBack = fpCallBack;
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	getTargetWindow
	//
	//	Summary		:	Get target window element from event
	//
	//==========================================================
	function getTargetWindow(e)
	{
		let oElement = e.target;
		
		if(!oElement)
			return null;
		
		for(let sID in frame.windows)
		{
			if(frame.windows[sID].element.contains(oElement))
				return frame.windows[sID];
		}
		return null;
	}
	
	
	//==========================================================
	//
	//	Function	:	startResizeWindow
	//
	//	Summary		:	Event handler of start resizing window
	//
	//==========================================================
	function startResizeWindow(e)
	{
		let oElement = e.target;
		
		// Target not found
		if(!oElement)
			return;
		
		// Find target window
		jaswin.currentWindow = getTargetWindow(e);
		if((!jaswin.currentWindow) || (jaswin.currentWindow == global))
		{
			jaswin.currentWindow = global;
			return;
		}
		
		if(oElement.style.cursor == "w-resize")
		{
			// For left direction
			iResizeMode = frame.RESIZE_LEFT;
		}
		else if(oElement.style.cursor == "e-resize")
		{
			// For right direction
			iResizeMode = frame.RESIZE_RIGHT;
		}
		else if(oElement.style.cursor == "n-resize")
		{
			// For top direction
			iResizeMode = frame.RESIZE_TOP;
		}
		else if(oElement.style.cursor == "s-resize")
		{
			// For bottom direction
			iResizeMode = frame.RESIZE_BOTTOM;
		}
		else if(oElement.style.cursor == "se-resize")
		{
			// For both direction
			iResizeMode = frame.RESIZE_BOTH;
		}
		else
		{
			jaswin.currentWindow = global;
			return;
		}
		
		global.element.addEventListener("mouseup", endResizeWindow, false);
		global.element.addEventListener("mousemove", resizeWindow, false);
	}
	
	
	//==========================================================
	//
	//	Function	:	resizeWindow
	//
	//	Summary		:	Event handler of resizing window
	//
	//==========================================================
	function resizeWindow(e)
	{
		let oWnd;
		let iCurX;
		let iCurY;
		let iParentWidth;
		let iParentHeight;
		let iTmpWidth;
		let iTmpHeight;
		let oSelection = window.getSelection();
		let oParentRect;
		
		oSelection.removeAllRanges();
		
		if((!jaswin.currentWindow) || (jaswin.currentWindow == global))
		{
			global.element.removeEventListener("mouseup", endResizeWindow, false);
			global.element.removeEventListener("mousemove", resizeWindow, false);
			jaswin.currentWindow = global;
			return;
		}
		oWnd = jaswin.currentWindow;
		
		oParentRect = oWnd.element.parentElement.getBoundingClientRect();
		iCurX = e.clientX - oParentRect.left;
		iCurY = e.clientY - oParentRect.top;
		
		// Get parent size
		iParentWidth = oWnd.element.parentElement.scrollWidth;
		iParentHeight = oWnd.element.parentElement.scrollHeight;
		
		// Prevent from running over
		iTmpWidth = oWnd.rect.width;
		
		if(iResizeMode == frame.RESIZE_LEFT)
		{
			// For left direction
			if(iCurX <= 0)
			{
				// Out of parent rect
				endResizeWindow(null);
				return;
			}
			
			if(iTmpWidth - (iCurX - oWnd.rect.left) <= oWnd.minWidth)
			{
				// Minimum size
				endResizeWindow(null);
				return;
			}
			
			iTmpWidth = iTmpWidth - (iCurX - oWnd.rect.left);
		}
		else if((iResizeMode == frame.RESIZE_RIGHT)
		|| (iResizeMode == frame.RESIZE_BOTH))
		{
			// For right or both direction
			if(iCurX <= oWnd.rect.left + oWnd.minWidth)
			{
				// Minimum size
				endResizeWindow(null);
				return;
			}
			
			if(iCurX >= iParentWidth - 2)
			{
				// Maximum size
				endResizeWindow(null);
				return;
			}
			
			iTmpWidth = iCurX - oWnd.rect.left;
		}
		
		iTmpHeight = oWnd.rect.height;
		
		if(iResizeMode == frame.RESIZE_TOP)
		{
			// For top direction
			if(iCurY <= 0)
			{
				// Out of parent rect
				endResizeWindow(null);
				return;
			}
			
			if(iTmpHeight - (iCurY - oWnd.rect.top) <= oWnd.minHeight)
			{
				// Minimum size
				endResizeWindow(null);
				return;
			}
			
			iTmpHeight = iTmpHeight - (iCurY - oWnd.rect.top);
		}
		else if((iResizeMode == frame.RESIZE_BOTTOM)
		|| (iResizeMode == frame.RESIZE_BOTH))
		{
			// For bottom or both direction
			if(iCurY <= oWnd.rect.top + oWnd.minHeight)
			{
				// Minimum size
				endResizeWindow(null);
				return;
			}
			if(iCurY >= iParentHeight - 2)
			{
				// Maximum size
				endResizeWindow(null);
				return;
			}
			
			iTmpHeight = iCurY - oWnd.rect.top;
		}
		
		// Set position
		if(iResizeMode == frame.RESIZE_LEFT)
			oWnd.setPosition(iCurX, oWnd.rect.top);
		else if(iResizeMode == frame.RESIZE_TOP)
			oWnd.setPosition(oWnd.rect.left, iCurY);
		
		// Resize window
		oWnd.setSize(iTmpWidth, iTmpHeight);
	}
	
	
	//==========================================================
	//
	//	Function	:	endResizeWindow
	//
	//	Summary		:	Event handler of end resizing window
	//
	//==========================================================
	function endResizeWindow(e)
	{
		global.element.removeEventListener("mouseup", endResizeWindow, false);
		global.element.removeEventListener("mousemove", resizeWindow, false);
	}
	
	
	//==========================================================
	//
	//	Function	:	startDragWindow
	//
	//	Summary		:	Event handler of start dragging window
	//
	//==========================================================
	function startDragWindow(e)
	{
		let oElement = e.target;
		
		// Target not found
		if(!oElement)
			return;
		
		// Find target window
		jaswin.currentWindow = getTargetWindow(e);
		if((!jaswin.currentWindow) || (jaswin.currentWindow == global))
		{
			jaswin.currentWindow = global;
			return;
		}
		
		// Escape close button
		if((oElement == jaswin.currentWindow.closeButton)
		|| (oElement == jaswin.currentWindow.icon))
			return;
		
		iMovingX = e.clientX;
		iMovingY = e.clientY;
		global.element.addEventListener("mouseup", endDragWindow, false);
		global.element.addEventListener("mousemove", dragWindow, false);
	}
	
	
	//==========================================================
	//
	//	Function	:	dragWindow
	//
	//	Summary		:	Event handler of dragging window
	//
	//==========================================================
	function dragWindow(e)
	{
		let oWnd;
		let iCurX;
		let iCurY;
		let iParentWidth;
		let iParentHeight;
		let iTmpRight;
		let iTmpBottom;
		let iNewX;
		let iNewY;
		let oSelection = window.getSelection();
		
		oSelection.removeAllRanges();
		
		if((!jaswin.currentWindow) || (jaswin.currentWindow == global))
		{
			global.element.removeEventListener("mouseup", endDragWindow, false);
			global.element.removeEventListener("mousemove", dragWindow, false);
			jaswin.currentWindow = global;
			return;
		}
		oWnd = jaswin.currentWindow;
		
		iCurX = e.clientX;
		iCurY = e.clientY;
		
		// Get parent size
		iParentWidth = oWnd.element.parentElement.scrollWidth;
		iParentHeight = oWnd.element.parentElement.scrollHeight;
		
		// Prevent from running over
		if((oWnd.rect.left + (iCurX - iMovingX) <= 0)
		&& (iCurX - iMovingX < 0))
		{
			endDragWindow(null);
			return;
		}
		if((oWnd.rect.top + (iCurY - iMovingY) <= 0)
		&& (iCurY - iMovingY < 0))
		{
			endDragWindow(null);
			return;
		}
		
		iTmpRight = oWnd.rect.left + (iCurX - iMovingX) + oWnd.rect.width;
		if((iTmpRight + 2 >= iParentWidth) && (iCurX - iMovingX > 0))
		{
			endDragWindow(null);
			return;
		}
		
		iTmpBottom = oWnd.rect.top + (iCurY - iMovingY) + oWnd.rect.height;
		if((iTmpBottom + 2 >= iParentHeight) && (iCurY - iMovingY > 0))
		{
			endDragWindow(null);
			return;
		}
		
		// Move window
		iNewX = oWnd.rect.left + (iCurX - iMovingX);
		iNewY = oWnd.rect.top + (iCurY - iMovingY);
		oWnd.setPosition(iNewX, iNewY);
		iMovingX = iCurX;
		iMovingY = iCurY;
	}
	
	
	//==========================================================
	//
	//	Function	:	endDragWindow
	//
	//	Summary		:	Event handler of end dragging window
	//
	//==========================================================
	function endDragWindow(e)
	{
		global.element.removeEventListener("mouseup", endDragWindow, false);
		global.element.removeEventListener("mousemove", dragWindow, false);
	}
	
}(jaswin.frame = jaswin.frame || {}));
